// -*- c-basic-offset: 4; indent-tabs-mode: nil -*-

/* ***** BEGIN LICENSE BLOCK *****
Version: MPL 1.1/GPL 2.0/LGPL 2.1

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Standard-Sitemap Navigator code.

The Initial Developer of the Original Code is the
Standard-Sitemap Team.

Contributor(s): (see license.txt)

Alternatively, the contents of this file may be used under the terms of
either the GNU General Public License Version 2 or later (the "GPL"), or
the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
in which case the provisions of the GPL or the LGPL are applicable instead
of those above. If you wish to allow use of your version of this file only
under the terms of either the GPL or the LGPL, and not to allow others to
use your version of this file under the terms of the MPL, indicate your
decision by deleting the provisions above and replace them with the notice
and other provisions required by the GPL or the LGPL. If you do not delete
the provisions above, a recipient may use your version of this file under
the terms of any one of the MPL, the GPL or the LGPL.
***** END LICENSE BLOCK ***** */

/* $Id: sslib.jsm 524 2011-09-02 21:56:02Z ssjools $
 *
 * Generic utility functions. Any function that has nothing to do with
 * Standard Sitemap and could be used in any extension should be
 * stored here.
 */


/* Note on use as a module:
 *
 * We don't export any symbols from this module, and instead just set
 * an entry in application storage, which you can retrieve with:
 * 
 * Application.storage.get("org.standard-sitemap.library", null)
 *
 * The advantage of this library being a module is that there's no
 * possibility of contamination from a window context, and that
 * contamination going unnoticed when used inadvertantly.  Instead,
 * you should get an error immediately. */

var EXPORTED_SYMBOLS = [];

var sslib = {
    Components : Components,

    extId      : 'extension@standard-sitemap.org',
    extName    : 'org.standard-sitemap.navigator',

    cout       : Components.classes["@mozilla.org/consoleservice;1"]
        .getService(Components.interfaces.nsIConsoleService),

    ioService  : Components.classes["@mozilla.org/network/io-service;1"]
        .getService(Components.interfaces.nsIIOService),

    oObserverService :
    Components.classes["@mozilla.org/observer-service;1"]
        .getService(Components.interfaces.nsIObserverService),

    initialize : function() {
        var sbService =
            Components.classes["@mozilla.org/intl/stringbundle;1"]
            .getService(Components.interfaces.nsIStringBundleService);
        this.oStringBundle = sbService.createBundle
        ("chrome://standard-sitemap.org/locale/standardsitemap.properties");
        this.oLanguageBundle = sbService.createBundle
        ("chrome://global/locale/languageNames.properties");
        this.oRegionBundle = sbService.createBundle
        ("chrome://global/locale/regionNames.properties");

        /* Get the current browsing mode. */
        // https://developer.mozilla.org/En/Supporting_private_browsing_mode
        try {
            var pbs = Components.classes["@mozilla.org/privatebrowsing;1"]
                .getService(Components.interfaces
                            .nsIPrivateBrowsingService);
            this.bIsPrivate = pbs.privateBrowsingEnabled;
        } catch(ex) {
            // Ignore exceptions in older versions of Firefox.
        }

        // Set up an observer to watch for changes in browsing
        // mode.
        var self = this;
        var privateWatcher = {
            observe : function(aSubject, aTopic, aData) {
                if (aTopic == "quit-application") {
                    self.oObserverService
                        .removeObserver(self, "quit-application");
                    self.oObserverService
                        .removeObserver(self, "private-browsing");
                    return;
                }
                if (aTopic == "private-browsing") {
                    if (aData == "enter")
                        self.enterPrivateBrowsing();
                    else if (aData == "exit")
                        self.exitPrivateBrowsing();
                    return;
                }
            },
        };
        this.oObserverService
            .addObserver(privateWatcher, "private-browsing", false);
        this.oObserverService
            .addObserver(privateWatcher, "quit-application", false);

        this.initPrefs();
    },

    dump : function(msg) {
        dump('[STDMAP] ' + msg + '\n');
        this.debug(msg);
    },
    
    debug : function(str) {
        if (this.prefs.debug) {
            this.cout.logStringMessage(this.stringSubst
                                       ("%1 (%3): %2", this.extName,
                                        str, new Date().getTime()));
        }
    },

    listMembers : function(obj) {
        if (!this.prefs.debug) return;
        var s = 'Members of ' + obj + ':\n';
        for (var key in obj) {
            s += '\t' + key + '\n';
        }
        this.debug(s);
    },

    makeURI : function(text) {
        if (text == null) return null;
        return this.ioService.newURI(text, 'us-ascii', null);
    },

    // Get xml:base of a node by looking at its ancestry for
    // xml:base.  'node' should be a DOM node.
    getBaseURI: function(oNode) {
        // TODO: When we drop support for @baseurl, the
        // implementation reduces to this single line.  Until
        // then, it is disabled, as it won't recognise @baseurl as
        // a back-up to @xml:base.

        //return this.makeURI(oNode.baseURI);

        // The following implementation uses @baseurl as a back-up
        // to @xml:base.

        // Have we worked this out before?
        if (oNode.sslib_base)
            return oNode.sslib_base;

        // Get the context for any URI specified at this node.
        var oContext = null;
        if ((oNode.parentNode &&
             (oContext = this.getBaseURI(oNode.parentNode),
              true)) ||
            (oNode.ownerDocument &&
             (oContext = this.makeURI(oNode.ownerDocument.documentURI),
              true))) {
            // This is a document node.  If it's not the root
            // element, its context is its document's URI;
            // otherwise, it is its parent's URI.

            // Does this node define xml:base (or sitemap:baseurl)?
            var sRaw;
            if (oNode.ownerDocument &&
                ((sRaw = oNode.getAttributeNS(this.xmlNS, 'base')) ||
                 (sRaw = this.getAttributeSM(oNode, 'baseurl')) ||
                 (this.issmns(oNode.namespaceURI) &&
                  (sRaw = oNode.getAttribute('baseurl'))))) {
                
                // Yes, so extract it, resolve it against the
                // context, and make an nsIURI object out of it.
                oNode.sslib_base = this.makeURI(oContext.resolve(sRaw));
            } else {
                // No, so the base of this element is the same as
                // its parent.
                oNode.sslib_base = oContext;
            }
        } else {
            // This is the whole document.  Just use its URI.
            oNode.sslib_base = this.makeURI(oNode.documentURI);
        }

        return oNode.sslib_base;
    },

    xmlNS : 'http://www.w3.org/XML/1998/namespace',

    xhtmlNS : 'http://www.w3.org/1999/xhtml',

    xlinkNS : 'http://www.w3.org/1999/xlink',

    sitemapNS : [
        'http://standard-sitemap.org/2007/ns',
        'http://www.standard-sitemap.org/version1beta/',
    ],

    xulNS : 'http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul',

    // Get an attribute in any of our namespaces.
    getAttributeSM : function(node, name) {
        for each (var cand in this.sitemapNS) {
            var val = node.getAttributeNS(cand, name);
            if (val)
                return val;
        }
        return null;
    },

    // Get the elements of a particular tag name in any of our
    // namespaces.  The first to produce a non-empty list is
    // returned.
    getElementsByTagNameSM : function(node, name) {
        for each (var cand in this.sitemapNS) {
            var elems = node.getElementsByTagNameNS(cand, name);
            if (elems != null && elems.length > 0)
                return elems;
        }
        return elems;
    },

    // Is this namespace one of ours?
    issmns : function(s) {
        if (!s) return false;
        var parts = s.trim().split(/\s+/);
        for each (var part in parts) {
            for each (var cand in this.sitemapNS) {
                if (cand == part)
                    return true;
            }
        }
        return false;
    },

    resolveInNode: function(oNode, sRaw) {
        if (!sRaw) return null
        var oBase = this.getBaseURI(oNode);
        return this.makeURI(oBase.resolve(sRaw));
    },

    getURIAttr: function(oNode, sName) {
        return this.resolveInNode(oNode, oNode.getAttribute(sName));
    },

    getURIAttrNS: function(oNode, sNamespace, sName) {
        return this.resolveInNode(oNode,
                                  oNode.getAttributeNS(sNamespace, sName));
    },

    objectToString : function(obj) {
        var s = '';
        var key;
        for (key in obj) {
            s += this.stringSubst("%1 : %2", key, obj[key]);
        }
        return s;
    },

    migratePrefs : function() {
        var names = [ 'modeWebmaster', 'referrer', 'debug',
                      'autoopen', 'urlbaricon', 'visitonselect',
                      'findonvisit', 'openonvisit', 'openonclick',
                      'selectonvisit', 'vertbar', 'showRoleItemsInTree',
                      'invertctrlforfocus', 'ctrlfornewtab',
                      'showUpButton', 'includeOptionalItems',
                      'installed' ];
        var oldBranch =
            Components.classes["@mozilla.org/preferences-service;1"]
            .getService(Components.interfaces.nsIPrefService)
            .getBranch("extensions.standardsitemap.");
        for each (var name in names) {
            // Only deal with old values that still exist.
            var oldType = oldBranch.getPrefType(name);
            if (oldType == 0) continue;
            var newType = this.prefBranch.getPrefType(name);
            switch (oldType) {
            case this.prefBranch.PREF_STRING:
                this.prefBranch.setCharPref(name,
                                            oldBranch.getCharPref(name));
                break;

            case this.prefBranch.PREF_INT:
                this.prefBranch.setIntPref(name,
                                           oldBranch.getIntPref(name));
                break;

            case this.prefBranch.PREF_BOOL:
                this.prefBranch.setBoolPref(name,
                                            oldBranch.getBoolPref(name));
                break;
            }
            oldBranch.clearUserPref(name);
        }
    },

    // Adds all prefs to a prefs object
    initPrefs : function() {
        this.prefBranch =
            Components.classes["@mozilla.org/preferences-service;1"]
            .getService(Components.interfaces.nsIPrefService)
            .getBranch(this.stringSubst("extensions.%1.", this.extName));
        this.prefBranch
            .QueryInterface(Components.interfaces.nsIPrefBranch2);
        this.prefBranch.addObserver("", this, false);

        //Make sure we don't leak
        //             window.addEventListener('unload', function(event) {
        //                 this.unregister();
        //             }, false);
        
        this.prefs = {};

        this.migratePrefs();

        var list = this.prefBranch.getChildList("", {})
            .toString().split(",");

        for (var i in list) {
            var name = list[i];
            var type = this.prefBranch.getPrefType(name);

            if (type == this.prefBranch.PREF_STRING) {
                this.prefs[name] = this.getCharPref(name);
            } else if (type == this.prefBranch.PREF_INT)  {
                this.prefs[name] = this.getIntPref(name);
            } else if (type == this.prefBranch.PREF_BOOL) {
                this.prefs[name] = this.getBoolPref(name);
            }
        }
        //this.setPref('autoopen', true); //set the specific pref
        //this.setPref('other', 'value'); //add new prefs
    },

    setPref : function(name, value) {
        var type = this.prefBranch.getPrefType(name);

        // undefined prefs, when adding new not specified in defaults
        if (type == 0) {
            if (typeof(value) == 'number') {
                type = 64;
            } else if (typeof(value) == 'string') {
                type = 32;
            } else if (typeof(value) == 'boolean') {
                type = 128;
            }
        }

        if (type == this.prefBranch.PREF_STRING) {
            this.setCharPref(name, value);
        } else if (type == this.prefBranch.PREF_INT)  {
            this.setIntPref(name, value);
        } else if (type == this.prefBranch.PREF_BOOL) {
            this.setBoolPref(name, value);
        }
    },

    //         unregister : function() {
    //             if(!this.prefBranch) return;
    //             this.prefBranch.removeObserver("", this);
    //         },

    observe : function(aSubject, aTopic, aData) {
        if(aTopic != "nsPref:changed") return;
        // HELP it works! but sometimes get error SSLib.dump is
        // not defined, check on other plaforms.

        // Out of scope sometimes? Have not seen it for a while
        // now, caveman 03-02-2009
        if (typeof(this.dump) == "function") {
            this.dump("Pref " + aData + " changed.");
            this.oObserverService
                .notifyObservers(null, "standard-sitemap-prefchanged",
                                 aData);
        }
    },

    getCharPref : function(branch) {
        return this.prefBranch.getCharPref(branch);
    },

    setCharPref : function(branch, value) {
        this.prefBranch.setCharPref(branch, value);
    },

    getBoolPref : function(branch) {
        return this.prefBranch.getBoolPref(branch);
    },

    setBoolPref : function(branch, value) {
        this.prefBranch.setBoolPref(branch, value);
    },

    getIntPref : function(branch) {
        return this.prefBranch.getIntPref(branch);
    },

    setIntPref : function(branch, value) {
        this.prefBranch.setIntPref(branch, value);
    },

    getMenuLimitPref : function() {
        var val = this.getIntPref("menu-limit");
        if (val < 5) return 5;
        return val;
    },

    savedContentPrefs :
    Components.classes["@mozilla.org/content-pref/service;1"]
        .getService(Components.interfaces.nsIContentPrefService),

    privateContentPrefs : { },

    bIsPrivate : false,

    enterPrivateBrowsing : function() {
        this.bIsPrivate = true;
    },

    exitPrivateBrowsing : function() {
        this.bIsPrivate = false;
        this.privateContentPrefs = { };
    },

    getContentPref : function(uri, key) {
        if (this.bIsPrivate) {
            var pkey = this.getPrivateContentPrefKey(uri, key);
            var val = this.privateContentPrefs[pkey];
            if (val) return val;
        }
        uri = this.fixSitePref(uri);
        return this.savedContentPrefs.getPref(uri, key);
    },

    setContentPref : function(uri, key, value) {
        if (this.bIsPrivate) {
            var pkey = this.getPrivateContentPrefKey(uri, key);
            this.privateContentPrefs[pkey] = value;
        } else {
            uri = this.fixSitePref(uri);
            this.savedContentPrefs.setPref(uri, key, value);
        }
    },

    removeContentPref : function(uri, key) {
        if (this.bIsPrivate) {
            var pkey = this.getPrivateContentPrefKey(uri, key);
            delete this.privateContentPrefs[pkey];
        } else {
            uri = this.fixSitePref(uri);
            this.savedContentPrefs.removePref(uri, key);
        }
    },

    checkLanguagePrefCallers : {},

    checkLanguagePref : function(caller) {
        if (caller in this.checkLanguagePrefCallers)
            return false;
        this.checkLanguagePrefCallers[caller] = true;
        return true;
    },

    getLanguagePref : function(uri) {
        if (!uri) return [];
        var name1 = 'org.standard-sitemap.languagepref';
        var name2 = 'org.standard-sitemap.sidebar.sidebarlanguagepref';
        var name3 = 'standardsitemap.sidebarlanguagepref';
        var pref = this.getContentPref(uri, name1);
        if (!pref)
            pref = this.getContentPref(uri, name2);
        if (!pref)
            pref = this.getContentPref(uri, name3);
        if (pref) return pref.split(',');
        return this.GetAcceptLanguages();
    },

    setLanguagePref: function(uri, newLang) {
        // There can be no preference for a site without a sitemap.
        if (!uri) return;

        this.checkLanguagePrefCallers = {};

        newLang = newLang.trim().toLowerCase();

        // Get the current preference, remove the language from the
        // sequence, then put it at the head.
        var pref = this.getLanguagePref();
        var pos = pref.indexOf(newLang);
        if (pos >= 0)
            pref = pref.slice(0, pos).concat(pref.slice(pos + 1));
        pref = [ newLang ].concat(pref);
        // TODO: If en is the new setting, move all en-* up with it,
        // in the order they previously appeared.  If en-gb is the new
        // setting, move all en-* in the same order, but put en-GB
        // first.

        var name1 = 'org.standard-sitemap.languagepref';

        // Record the new preference.
        this.setContentPref(uri, name1, pref.toString());
        this.debug('New language preference ' + pref +
                   ' for ' + uri.spec + " " + typeof(pref));
    },

    clearLanguagePref: function(uri) {
        // There can be no preference for a site without a sitemap.
        if (!uri) return;
        var name1 = 'org.standard-sitemap.languagepref';
        var name2 = 'org.standard-sitemap.sidebar.sidebarlanguagepref';
        var name3 = 'standardsitemap.sidebarlanguagepref';
        this.removeContentPref(uri, name1);
        this.removeContentPref(uri, name2);
        this.removeContentPref(uri, name3);
    },

    getPrivateContentPrefKey : function(uri, key) {
        if (!uri) return '>' + key;
        return uri.spec + '>' + key;
    },

    // The Content Preference Service uses a narrow definition of
    // 'site'.  In fact, it's handles host preferences, but we want
    // preferences per URI, not per host.  This function translates
    // one nsIURI into another, incorporating the path component into
    // a questionable domainname appended to the original host.
    fixSitePref : function(loc) {
        if (loc.asciiSpec == 'null:') return loc;
        //this.debug('original: ' + loc.spec);
        var asciiPath = loc.asciiSpec.substring(loc.prePath.length);
        var r = loc.scheme + '://';
        if (loc.userPass) r += loc.userPass + '@';
        r += loc.host;
        var skip = 0;
        for each (var c in asciiPath) {
            if (skip > 0) {
                skip--;
                r += c;
                continue;
            }

            switch (c) {
            case 'a': case 'b': case 'c': case 'd': case 'e':
            case 'f': case 'g': case 'h': case 'i': case 'j':
            case 'k': case 'l': case 'm': case 'n': case 'o':
            case 'p': case 'q': case 'r': case 's': case 't':
            case 'u': case 'v': case '0': case '1': case '2':
            case '3': case '4': case '5': case '6': case '7':
            case '8': case '9': 
                r += c;
                break;

            case 'w': case 'x': case 'y': case 'z':
                r += 'w' + c.toLowerCase();
                break;

            case 'A': case 'B': case 'C': case 'D': case 'E':
            case 'F': case 'G': case 'H': case 'I': case 'J':
            case 'K': case 'L': case 'M': case 'N': case 'O':
            case 'P': case 'Q': case 'R': case 'S': case 'T':
            case 'U': case 'V': case 'W': case 'X': case 'Y': case 'Z':
                r += 'x' + c.toLowerCase();
                break;

            case '/':
                r += '.';
                break;

            case '%':
                r += 'y';
                skip = 2;
                break;

            default:
                r += 'z' + c.charCodeAt(0).toString(16);
                break;
            }
        }
        r += '.illegal';
        if (loc.port > -1)
            r += ':' + loc.port;
        r += '/';
        //this.debug('transformed: ' + r);
        return this.makeURI(r);
    },

    // Work out which URI, if any, to use as a referrer.
    //
    // 'mode' is an array of preferred options.  Each item is
    // tested until one is found to have meaning.
    //
    // 'rootURI' should be the URI of the root sitemap file.
    //
    // 'uiURI' should be the URI of the sitemap file containing
    // the node in question.
    //
    // 'parentUiURI' should be the destination URI of the nearest
    // ancester node of the node in question that specifies a
    // destination.
    //
    // 'currentURI' should be the URI of the current location.
    getReferrer : function(mode, rootURI, parentUiURI, uiURI,
                           currentURI) {
        var result = null;
        for each (var modeItem in mode) {
            switch (modeItem) {
            case 'root':
                if (!rootURI) continue;
                return rootURI;

            case 'map':
                if (!uiURI) continue;
                return uiURI;

            case 'page':
                if (!currentURI) continue;
                return currentURI;

            case 'parent':
                if (!parentUiURI) continue;
                return parentUiURI;

            case 'none':
                return null;

            case 'user':
                switch (this.getCharPref('referrer')) {
                case 'sitemap':
                    if (!rootURI) continue;
                    return rootURI;
                case 'current':
                    if (!currentURI) continue;
                    return currentURI;
                default:
                    return null;
                }

            default:
                return;
            }
        }
        switch (this.getCharPref('referrer')) {
        case 'sitemap':
            return rootURI;
        case 'current':
            return currentURI;
        default:
            return null;
        }
    },

    getLanguageName : function(s) {
        try {
            if (s in this.languages)
                return this.languages[s].nativeName;
            return this.oLanguageBundle.GetStringFromName(s);
        } catch (e) {
            return "Unknown";
        }
    },

    getRegionName : function(s) {
        try {
            return this.oRegionBundle.GetStringFromName(s);
        } catch (e) {
            return "Unknown";
        }
    },

    nameLanguageCode : function(s) {
        var parts = s.trim().toLowerCase().split(/[-_.]/);
        //this.debug('code ' + s + ' has ' + parts.length + ' parts');
        // TODO: Localize this string.
        if (parts.length == 0) return "Unknown";
        //this.debug("looking for lang " + parts[0]);
        var r = "" + this.getLanguageName(parts[0]);
        if (parts.length == 1)
            return r;
        //this.debug("looking for region " + parts[1]);
        r += " (" + this.getRegionName(parts[1]) + ")";
        return r;
    },

    TranslateString: function(sName, sVar1, sVar2, sVar3)
    {
        var sResult = "";
        if(this.oStringBundle)
        {
            sResult  = this.oStringBundle.GetStringFromName(sName);
            if (sVar1 || sVar1 == 0)
                sResult = sResult.replace(/%1/g, sVar1);
            if (sVar2 || sVar2 == 0)
                sResult = sResult.replace(/%2/g, sVar2);
            if (sVar3 || sVar3 == 0)
                sResult = sResult.replace(/%3/g, sVar3);
        }
        return sResult;
    },

    getEnvVar : function(name) {
        return this.Components.classes["@mozilla.org/process/environment;1"]
            .getService(this.Components.interfaces.nsIEnvironment)
            .get(name);
    },

    setEnvVar : function(name, value) {
        return this.Components.classes["@mozilla.org/process/environment;1"]
            .getService(this.Components.interfaces.nsIEnvironment)
            .set(name, value);
    },

    getExtensionFolder : function() {
        return this.Components.classes["@mozilla.org/extensions/manager;1"]
            .getService(this.Components.interfaces.nsIExtensionManager)
            .getInstallLocation(this.extId)
            .getItemLocation(this.extId);
    },

    msgBox : function(title, text) {
        this.Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
            .getService(this.Components.interfaces.nsIPromptService)
            .alert(window, title, text);
    },

    loadScript : function(url) {
        var jsLoader =
            this.Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
            .getService(this.Components.interfaces.mozIJSSubScriptLoader);
        jsLoader.loadSubScript(url);
    },

    isVisitied : function(oURI) {
        oGlobalHistory =
            this.Components.classes["@mozilla.org/browser/global-history;2"]
            .getService(this.Components.interfaces.nsIGlobalHistory2);
        return oURI ? oGlobalHistory.isVisited(oURI) : false;
    },

    // Get the accept_languages preferences as an array of items,
    // since this might be a localized pref we need to try and get
    // it as a complex first.
    GetAcceptLanguages: function()
    {
        var oALPref =
            this.Components.classes["@mozilla.org/preferences-service;1"]
            .getService(this.Components.interfaces.nsIPrefService)
            .getBranch("intl.");
        var aResult = new Array();
        var sResult = "";
        try {
            sResult =
                oALPref.getComplexValue("accept_languages",
                                        this.Components.interfaces
                                        .nsIPrefLocalizedString).data;
        } catch (e) {
            sResult = "";
        }
        sResult = oALPref.getCharPref("accept_languages");
        if (sResult != "") aResult = sResult.split(",");
        return aResult;
    },

    // Saves 'content' to file 'filepath'. Note that filepath
    // needs to be a real file path, not a chrome path.
    saveToFile : function(filepath, content) {
        var file = this.getFile(filepath);

        if (!file.exists()) {
            file.create(this.Components.interfaces
                        .nsIFile.NORMAL_FILE_TYPE, 420);
        }
        var outputStream = this.Components.classes
        ["@mozilla.org/network/file-output-stream;1"]
            .createInstance(this.Components.interfaces.nsIFileOutputStream);

        outputStream.init( file, 0x04 | 0x08 | 0x20, 420, 0 );
        var result = outputStream.write( content, content.length );
        outputStream.close();
    },

    // Returns the content of 'file' as string. The 'file'
    // parameter should be a nsIFile object.
    getFileContents : function(file) {
        var fstream = this.Components.classes
        ["@mozilla.org/network/file-input-stream;1"]
            .createInstance(this.Components.interfaces.nsIFileInputStream);
        var sstream =
            this.Components.classes["@mozilla.org/scriptableinputstream;1"]
            .createInstance(this.Components.interfaces
                            .nsIScriptableInputStream);
        fstream.init(file, -1, 0, 0);
        sstream.init(fstream);

        var data = '';

        var str = sstream.read(4096);
        while (str.length > 0) {
            data += str;
            str = sstream.read(4096);
        }
        fstream.close();
        return data;
    },

    // Returns the content of 'file' as string. The 'file'
    // parameter should be a nsIFile object.
    getBinaryFileContents : function(file) {
        var fstream = this.Components.classes
        ["@mozilla.org/network/file-input-stream;1"]
            .createInstance(this.Components.interfaces.nsIFileInputStream);
        fstream.init(file, -1, -1, false);

        var bstream = this.Components.classes
        ["@mozilla.org/binaryinputstream;1"]
            .createInstance(this.Components.interfaces
                            .nsIBinaryInputStream);
        bstream.setInputStream(fstream);

        var bytes = bstream.readBytes(bstream.available());
        fstream.close();
        return bytes;
    },

    startProcess: function(filename, args) {
        var file = this.getFile(filename);

        args = args ? args : [];

        if (file.exists()) {
            var nsIProcess =
                this.Components.classes["@mozilla.org/process/util;1"]
                .getService(this.Components.interfaces.nsIProcess);
            nsIProcess.init(file);
            nsIProcess.run(false, args, args.length);
        } else {
            throw Error(this.stringSubst("File '%1' does not exist!",
                                         filename));
        }
    },

    // Simulates a double click on the file in question.
    launchFile : function(filepath) {
        var f = this.getFile(filepath);
        f.launch();
    },

    // Gets a local file reference, return the interface
    // nsILocalFile.
    getFile : function(filepath) {
        var f = this.Components.classes["@mozilla.org/file/local;1"]
            .createInstance(this.Components.interfaces.nsILocalFile);
        f.initWithPath(filepath);
        return f;
    },

    filePicker : function(name) {
        var nsIFilePicker = this.Components.interfaces.nsIFilePicker;
        var FilePicker =
            this.Components.classes["@mozilla.org/filepicker;1"]
            .createInstance(nsIFilePicker);
        FilePicker.init(window, name, nsIFilePicker.modeOpen);
        FilePicker.appendFilters(nsIFilePicker.filterXML);
        var ReturnFP = FilePicker.show();
        if (ReturnFP == nsIFilePicker.returnOK ||
            ReturnFP == nsIFilePicker.returnReplace) {
            var Picker = Array();
            Picker['file'] = FilePicker.file;
            Picker['path'] = FilePicker.file.path;
            Picker['fileURI'] = this.ioService.newFileURI(Picker['file']);
            return Picker;
        } else {
            return false;
        }
    },

    // Returns all elements that match the query sent in. The root
    // used in the query is the window.content.document, so this
    // will only work for html content.
    //         xpath : function(query, window) {
    //             return window.content.document.evaluate
    //             (query, window.content.document, null,
    //              XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE, null);
    //         },

    // Adjust the set of classes on an element.
    setClassScale : function(sElemPath, sAttrName, sAttrNS, fNSresolver,
                             sBase, iMin, iMax, iVal, doc, window) {
        var elem = doc.evaluate(sElemPath, doc, fNSresolver,
                                window.XPathResult.FIRST_ORDERED_NODE_TYPE,
                                null);
        if (!elem) return;
        elem = elem.singleNodeValue;
        if (!elem) return;
        var attrVal = sAttrNS ?
            elem.getAttributeNS(sAttrNS, sAttrName) :
            elem.getAttribute(sAttrName);

        // Get the current set of classes.
        var classes = { };
        if (attrVal) {
            for each (var s in attrVal.trim().split(/\s+/)) {
                classes[s] = true;
            }
        }

        // Adjust the set.
        for (var i = iMin; i < iVal; i++) {
            delete classes[sBase + "-under-" + i];
            classes[sBase + "-over-" + i] = true;
        }
        delete classes[sBase + "-under-" + iVal];
        delete classes[sBase + "-over-" + iVal];
        for (var i = iVal + 1; i <= iMax; i++) {
            classes[sBase + "-under-" + i] = true;
            delete classes[sBase + "-over-" + i];
        }

        // Write the classes back to the element.
        attrVal = "";
        for (var i in classes) {
            if (classes[i])
                attrVal += " " + i;
        }
        if (sAttrNS)
            elem.setAttributeNS(sAttrNS, sAttrName, attrVal);
        else
            elem.setAttribute(sAttrName, attrVal);
    },

    setStatusBarText : function(obj, window) {
        var text;
        if (obj.getAttribute) { // This is an element.
            text = obj.getAttribute('statustext');
        } else {
            text = obj;
        }

        if (text) {
            var statusbar = window.top.document
                .getElementById('statusbar-display');

            if (statusbar) {
                statusbar.oldtext = statusbar.getAttribute('label');
                statusbar.setAttribute('label', text);
            }
        }
    },

    clearStatusBarText : function(window) {
        var statusbar =
            window.top.document.getElementById('statusbar-display');

        if (statusbar) {
            statusbar.setAttribute('label', statusbar.oldtext ?
                                   statusbar.oldtext : '');
        }
    },

    getStatusBarText : function(window) {
        var statusbar =
            window.top.document.getElementById('statusbar-display');

        if (statusbar) {
            return statusbar.getAttribute('label');
        }
        return '';
    },
    
    GetString: function(sPropertyName, document)
    {
        var oStrings = document.getElementById("standardsitemap-strings");
        if (oStrings) return oStrings.getString(sPropertyName);
        return "Error unknown property: " + sPropertyName;
    },
    
    GetAppVersion: function()
    {
        return this.Components.classes["@mozilla.org/xre/app-info;1"]
            .getService(this.Components.interfaces.nsIXULAppInfo).version;
    },
    
    // Generates a probably unique id in a 8 long HEX format
    // (09AFFA90) (there is a chance that it might not be unique I
    // guess...)
    GenerateID: function()
    {
        return ((((1+Math.random())*0x10000)|0).toString(16).substring(1) +
                (((1+Math.random())*0x10000)|0).toString(16).substring(1))
            .toUpperCase();
    },
    
    ReadCSS: function(window, sURI)
    {
        try
        {
            var oRequest = new window.XMLHttpRequest();
            oRequest.open("GET", sURI, false);
            oRequest.overrideMimeType("text/css");
            oRequest.send(null);

            if (oRequest.status != 404)
            {
                if (oRequest.responseText) return oRequest.responseText;
            }
        } catch (e) {
            this.debug('Obtaining CSS from ' + sURI + ': ' + e);
            return "";
        }
        return "";
    },
    
    IsLastWindow: function()
    {
        var oWW = this.Components
            .classes["@mozilla.org/embedcomp/window-watcher;1"]
            .getService(this.Components.interfaces.nsIWindowWatcher);
        var oWEnum = oWW.getWindowEnumerator();
        var iCount = 0;
        var bResult = true;
        while (oWEnum.hasMoreElements)
        {
            iCount ++;
            oWEnum.getNext();
            if (iCount > 10) // Fail save...
            {
                bResult = false;
                break;
            }
        }
        if (iCount <= 1) bResult = true;
        return bResult;
    },

    // Returns true if the string contains the substring str.
    stringContains : function (s, str) {
        return s.indexOf(str) != -1;
    },

    stringTrim : function(s) {
        return s.replace(/^\s*|\s*$/gi, '');
    },

    stringStartsWith : function(b, s) {
        return (b.length >= s.length && b.substr(0, s.length) == s);
    },

    stringEndsWith : function(b, s) {
        return (b.length >= s.length && b.substr(b.length - s.length) == s);
    },

    // Inserts the arguments into the string. E.g. if the string is
    // "Hello %1" then stringSubst("Hello %1", 'johnny') would return
    // "Hello johnny"
    stringSubst : function() {
        var s = arguments[0];
        for (var i = 1; i < arguments.length; i++) {
            var nr = "%" + i;
            var repl;
            if (arguments[i] == null) {
                repl = "null";
            } else if (arguments[i] == undefined) {
                repl = "undefined";
            } else {
                repl = arguments[i];
            }
            s = s.replace(new RegExp(nr, "g"), repl);
        }
        return s;
    },

    xulHide : function(b) {
        b.style.display = 'none';
    },

    xulShow : function(b) { 
        b.style.display = '';
    },

    xulEnable : function(b) {
        if (b)
            b.setAttribute('disabled', false);
    },

    xulDisable : function(b) {
        if (b)
            b.setAttribute('disabled', true);
    },

    xulAddClass : function(b, Class) {
        var current = b.getAttribute('class') + '';
        var classes = current.split(' ');
        for (var i = 0; i < classes.length; i++)
            if (classes[i] == Class)
                return;
        b.setAttribute('class', current + ' ' + Class);
    },

    xulRemoveClass : function(b, Class) {
        var current = b.getAttribute('class') + '';
        var classes = current.split(' ');
        var newclass = '';
        for each (var name in classes) {
            if (name != Class) {
                if (newclass != '')
                    newclass += ' ';
                newclass += name;
            }
        }
        b.setAttribute('class', newclass);
    },

    xulAddProperty : function(b, Property) {
        var current = b.getAttribute('properties') + '';
        var properties = current.split(' ');
        for each (var p in properties) {
            if (p == Property)
                return;
        }
        b.setAttribute('properties', current + ' ' + Property);
    },

    xulRemoveProperty : function(b, Property) {
        var current = b.getAttribute('properties') + '';
        var properties = current.split(' ');
        var newproperties = '';
        for each (var name in properties) {
            if (name != Property) {
                if (newproperties != '')
                    newproperties += ' ';
                newproperties += name;
            }
        }
        b.setAttribute('properties', newproperties);
    },

    xulRemoveAllChildren : function(b) {
        while (b.firstChild) {
            b.removeChild(b.firstChild);
        }
    },


    xmlGetElementById : function (b, myID) {
        return b.evaluate('//*[@xml:id="'+myID+'"]', b,
                          function () { return this.xmlNS; },
                          XPathResult.FIRST_ORDERED_NODE_TYPE,
                          null).singleNodeValue;
    },

    Map : function(size, fHash, fEquals) {
        if (typeof size != "integer")
            size = 271;
        if (typeof fHash != "function")
            fHash = function(obj) {
                var r = 0;
                var s = toString(obj);
                var len = s.length;
                for (var i = 0; i < len; i++)
                    r += s.charCodeAt(i);
                return r;
            };
        if (typeof fEquals != "function")
            fEquals = function(obj1, obj2) {
                return obj1 == obj2;
            };
        this.fHash = fHash;
        this.fEquals = fEquals;
        this.size = size;
        this.clear();
    },

    // TODO: Does JavaScript have any sort of general Map class,
    // i.e. mapping object to object?  Seems that an associative array
    // can only handle strings, i.e. it converts its keys to strings
    // beforehand.  If there is a native Map class, use that instead
    // of this.
    newMap : function(size, fHash, fEquals) {
        return new (this.Map)(size, fHash, fEquals);
    },
};



// List of languages screenscraped from
// <http://en.wikipedia.org/wiki/List_of_ISO_639-2_codes>.  If you
// need to make a lot of changes to this section consider running the
// screenscraping script again instead of changing manually. You MUST
// save this file as utf-8 so the strange characters don't get messed
// up.
sslib.languages = {
    abk : { englishName : "Abkhaz", nativeName : "Аҧсуа", flag : ""},
    ab : { englishName : "Abkhaz", nativeName : "Аҧсуа", flag : ""},
    ace : { englishName : "Acehnese, Achinese", nativeName : "Aceh", flag : ""},
    ach : { englishName : "Acoli", nativeName : "Acoli", flag : ""},
    ada : { englishName : "Adangme", nativeName : "adangbɛ", flag : ""},
    ady : { englishName : "Adyghe", nativeName : "адыгэбзэ", flag : ""},
    aar : { englishName : "Afar", nativeName : "Afaraf", flag : ""},
    aa : { englishName : "Afar", nativeName : "Afaraf", flag : ""},
    afh : { englishName : "Afrihili", nativeName : "Afrihili", flag : ""},
    afr : { englishName : "Afrikaans", nativeName : "Afrikaans", flag : ""},
    af : { englishName : "Afrikaans", nativeName : "Afrikaans", flag : ""},
    afa : { englishName : "Afro-Asiatic (Other)", nativeName : "Afro-Asiatic (Other)", flag : ""},
    ain : { englishName : "Ainu", nativeName : "アイヌ イタㇰ(イタッㇰ) or ainu itak", flag : ""},
    aka : { englishName : "Akan", nativeName : "Akan", flag : ""},
    ak : { englishName : "Akan", nativeName : "Akan", flag : ""},
    akk : { englishName : "Akkadian", nativeName : "akkadû", flag : ""},
    alb : { englishName : "Albanian", nativeName : "Shqip", flag : ""},
    sqi : { englishName : "Albanian", nativeName : "Shqip", flag : ""},
    sq : { englishName : "Albanian", nativeName : "Shqip", flag : ""},
    ale : { englishName : "Aleut", nativeName : "Aleut", flag : ""},
    alg : { englishName : "Algonquian languages", nativeName : "Algonquian languages", flag : ""},
    tut : { englishName : "Altaic (Other)", nativeName : "Altaic (Other)", flag : ""},
    amh : { englishName : "Amharic", nativeName : "አማርኛ", flag : ""},
    am : { englishName : "Amharic", nativeName : "አማርኛ", flag : ""},
    anp : { englishName : "Angika", nativeName : "Angika", flag : ""},
    apa : { englishName : "Apache languages", nativeName : "Apache languages", flag : ""},
    ara : { englishName : "Arabic", nativeName : "العربية", flag : ""},
    ar : { englishName : "Arabic", nativeName : "العربية", flag : ""},
    arg : { englishName : "Aragonese", nativeName : "Aragonés", flag : ""},
    an : { englishName : "Aragonese", nativeName : "Aragonés", flag : ""},
    arc : { englishName : "Aramaic", nativeName : "ܐܪܡܝܐ", flag : ""},
    arp : { englishName : "Arapaho", nativeName : "Hinóno'eitíít", flag : ""},
    arn : { englishName : "Araucanian", nativeName : "mapudungun", flag : ""},
    arw : { englishName : "Arawak", nativeName : "Arawak", flag : ""},
    arm : { englishName : "Armenian", nativeName : "Հայերեն լեզու", flag : ""},
    hye : { englishName : "Armenian", nativeName : "Հայերեն լեզու", flag : ""},
    hy : { englishName : "Armenian", nativeName : "Հայերեն լեզու", flag : ""},
    rup : { englishName : "Aromanian", nativeName : "Armãneashce", flag : ""},
    art : { englishName : "Artificial (Other)", nativeName : "Artificial (Other)", flag : ""},
    asm : { englishName : "Assamese", nativeName : "অসমীয়া", flag : ""},
    as : { englishName : "Assamese", nativeName : "অসমীয়া", flag : ""},
    ast : { englishName : "Asturian", nativeName : "asturianu", flag : ""},
    ath : { englishName : "Athapascan languages", nativeName : "Athapascan languages", flag : ""},
    aus : { englishName : "Australian languages", nativeName : "Australian languages", flag : ""},
    bar : { englishName : "Austro-Bavarian", nativeName : "boarisch", flag : ""},
    map : { englishName : "Austronesian (Other)", nativeName : "Austronesian (Other)", flag : ""},
    ava : { englishName : "Avaric", nativeName : "авар мацӀ", flag : ""},
    av : { englishName : "Avaric", nativeName : "авар мацӀ", flag : ""},
    ave : { englishName : "Avestan", nativeName : "avesta", flag : ""},
    ae : { englishName : "Avestan", nativeName : "avesta", flag : ""},
    awa : { englishName : "Awadhi", nativeName : "अवधी", flag : ""},
    aym : { englishName : "Aymara", nativeName : "aymar aru", flag : ""},
    ay : { englishName : "Aymara", nativeName : "aymar aru", flag : ""},
    aze : { englishName : "Azerbaijani", nativeName : "Azərbaycanca", flag : ""},
    az : { englishName : "Azerbaijani", nativeName : "Azərbaycanca", flag : ""},
    ban : { englishName : "Balinese", nativeName : "Basa Bali", flag : ""},
    bat : { englishName : "Baltic (Other)", nativeName : "Baltic (Other)", flag : ""},
    bal : { englishName : "Baluchi", nativeName : "بلوچی", flag : ""},
    bam : { englishName : "Bambara", nativeName : "bamanankan", flag : ""},
    bm : { englishName : "Bambara", nativeName : "bamanankan", flag : ""},
    bai : { englishName : "Bamileke languages", nativeName : "Bamileke languages", flag : ""},
    bad : { englishName : "Banda", nativeName : "Banda", flag : ""},
    bnt : { englishName : "Bantu (Other)", nativeName : "Bantu (Other)", flag : ""},
    bas : { englishName : "Basa", nativeName : "ɓasaá", flag : ""},
    bak : { englishName : "Bashkir", nativeName : "башҡорт теле", flag : ""},
    ba : { englishName : "Bashkir", nativeName : "башҡорт теле", flag : ""},
    baq : { englishName : "Basque", nativeName : "euskara", flag : ""},
    eus : { englishName : "Basque", nativeName : "euskara", flag : ""},
    eu : { englishName : "Basque", nativeName : "euskara", flag : ""},
    btk : { englishName : "Batak (Indonesia)", nativeName : "Batak (Indonesia)", flag : ""},
    bej : { englishName : "Beja", nativeName : "بداوية", flag : ""},
    bel : { englishName : "Belarusian", nativeName : "Беларуская мова", flag : ""},
    be : { englishName : "Belarusian", nativeName : "Беларуская мова", flag : ""},
    bem : { englishName : "Bemba", nativeName : "ichiBemba", flag : ""},
    ben : { englishName : "Bengali", nativeName : "বাংলা", flag : ""},
    bn : { englishName : "Bengali", nativeName : "বাংলা", flag : ""},
    ber : { englishName : "Berber (Other)", nativeName : "Berber (Other)", flag : ""},
    bho : { englishName : "Bhojpuri", nativeName : "भोजपुरी", flag : ""},
    bih : { englishName : "Bihari", nativeName : "Bihari", flag : ""},
    bh : { englishName : "Bihari", nativeName : "Bihari", flag : ""},
    bik : { englishName : "Bikol languages", nativeName : "Bikol languages", flag : ""},
    bin : { englishName : "Bini", nativeName : "Edo", flag : ""},
    bis : { englishName : "Bislama", nativeName : "Bislama", flag : ""},
    bi : { englishName : "Bislama", nativeName : "Bislama", flag : ""},
    byn : { englishName : "Blin", nativeName : "ብሊና", flag : ""},
    bos : { englishName : "Bosnian", nativeName : "bosanski jezik", flag : ""},
    bs : { englishName : "Bosnian", nativeName : "bosanski jezik", flag : ""},
    bra : { englishName : "Brij Bhasha", nativeName : "ब्रज भाषा", flag : ""},
    bre : { englishName : "Breton", nativeName : "brezhoneg", flag : ""},
    br : { englishName : "Breton", nativeName : "brezhoneg", flag : ""},
    bug : { englishName : "Buginese", nativeName : "ᨅᨔ ᨕᨘᨁᨗ", flag : ""},
    bul : { englishName : "Bulgarian", nativeName : "български език", flag : ""},
    bg : { englishName : "Bulgarian", nativeName : "български език", flag : ""},
    bua : { englishName : "Buriat", nativeName : "буряад хэлэн", flag : ""},
    bur : { englishName : "Burmese", nativeName : "ျမန္မာစာ", flag : ""},
    mya : { englishName : "Burmese", nativeName : "ျမန္မာစာ", flag : ""},
    my : { englishName : "Burmese", nativeName : "ျမန္မာစာ", flag : ""},
    cad : { englishName : "Caddo", nativeName : "Hasí:nay", flag : ""},
    car : { englishName : "Carib", nativeName : "Carib", flag : ""},
    cat : { englishName : "Catalan", nativeName : "català", flag : ""},
    ca : { englishName : "Catalan", nativeName : "català", flag : ""},
    cau : { englishName : "Caucasian (Other)", nativeName : "Caucasian (Other)", flag : ""},
    ceb : { englishName : "Cebuano", nativeName : "Sinugboanon", flag : ""},
    cel : { englishName : "Celtic (Other)", nativeName : "Celtic (Other)", flag : ""},
    cai : { englishName : "Central American Indian (Other)", nativeName : "Central American Indian (Other)", flag : ""},
    chg : { englishName : "Chagatai", nativeName : "جغتای", flag : ""},
    cmc : { englishName : "Chamic languages", nativeName : "Chamic languages", flag : ""},
    cha : { englishName : "Chamorro", nativeName : "Chamoru", flag : ""},
    ch : { englishName : "Chamorro", nativeName : "Chamoru", flag : ""},
    che : { englishName : "Chechen", nativeName : "нохчийн мотт", flag : ""},
    ce : { englishName : "Chechen", nativeName : "нохчийн мотт", flag : ""},
    chr : { englishName : "Cherokee", nativeName : "ᏣᎳᎩ", flag : ""},
    chy : { englishName : "Cheyenne", nativeName : "Tsêhést", flag : ""},
    chb : { englishName : "Chibcha", nativeName : "Chibcha", flag : ""},
    nya : { englishName : "Chichewa", nativeName : "chiCheŵa", flag : ""},
    ny : { englishName : "Chichewa", nativeName : "chiCheŵa", flag : ""},
    chi : { englishName : "Chinese", nativeName : "中文", flag : ""},
    zho : { englishName : "Chinese", nativeName : "中文", flag : ""},
    zh : { englishName : "Chinese", nativeName : "中文", flag : ""},
    chn : { englishName : "Chinook jargon", nativeName : "Chinook jargon", flag : ""},
    cho : { englishName : "Choctaw", nativeName : "Chahta Anumpa", flag : ""},
    chu : { englishName : "Church Slavonic", nativeName : "ѩзыкъ словѣньскъ", flag : ""},
    cu : { englishName : "Church Slavonic", nativeName : "ѩзыкъ словѣньскъ", flag : ""},
    chk : { englishName : "Chuukese", nativeName : "Chuukese", flag : ""},
    chv : { englishName : "Chuvash", nativeName : "чӑваш чӗлхи", flag : ""},
    cv : { englishName : "Chuvash", nativeName : "чӑваш чӗлхи", flag : ""},
    nwc : { englishName : "Classical Newari", nativeName : "Classical Newari", flag : ""},
    syc : { englishName : "Classical Syriac", nativeName : "Classical Syriac", flag : ""},
    cop : { englishName : "Coptic", nativeName : "ⲙⲉⲧⲛ̀ⲣⲉⲙⲛ̀ⲭⲏⲙⲓ", flag : ""},
    cor : { englishName : "Cornish", nativeName : "Kernewek", flag : ""},
    kw : { englishName : "Cornish", nativeName : "Kernewek", flag : ""},
    cos : { englishName : "Corsican", nativeName : "corsu", flag : ""},
    co : { englishName : "Corsican", nativeName : "corsu", flag : ""},
    cre : { englishName : "Cree", nativeName : "ᓀᐦᐃᔭᐍᐏᐣ", flag : ""},
    cr : { englishName : "Cree", nativeName : "ᓀᐦᐃᔭᐍᐏᐣ", flag : ""},
    mus : { englishName : "Creek", nativeName : "Maskoki", flag : ""},
    crp : { englishName : "Creoles and Pidgins (Other)", nativeName : "Creoles and Pidgins (Other)", flag : ""},
    cpe : { englishName : "Creoles and Pidgins, English-based (Other)", nativeName : "Creoles and Pidgins, English-based (Other)", flag : ""},
    cpf : { englishName : "Creoles and Pidgins, French-based (Other)", nativeName : "Creoles and Pidgins, French-based (Other)", flag : ""},
    cpp : { englishName : "Creoles and Pidgins, Portuguese-based (Other)", nativeName : "Creoles and Pidgins, Portuguese-based (Other)", flag : ""},
    crh : { englishName : "Crimean Tatar (Crimean Turkish)", nativeName : "qırımtatar tili, къырымтатар тили", flag : ""},
    scr : { englishName : "Croatian", nativeName : "hrvatski jezik", flag : ""},
    hrv : { englishName : "Croatian", nativeName : "hrvatski jezik", flag : ""},
    hr : { englishName : "Croatian", nativeName : "hrvatski jezik", flag : ""},
    cus : { englishName : "Cushitic (Other)", nativeName : "Cushitic (Other)", flag : ""},
    cze : { englishName : "Czech", nativeName : "čeština (substantive)", flag : ""},
    ces : { englishName : "Czech", nativeName : "čeština (substantive)", flag : ""},
    cs : { englishName : "Czech", nativeName : "čeština (substantive)", flag : ""},
    dak : { englishName : "Dakota", nativeName : "Lakhota", flag : ""},
    dan : { englishName : "Danish", nativeName : "dansk", flag : ""},
    da : { englishName : "Danish", nativeName : "dansk", flag : ""},
    dar : { englishName : "Dargwa", nativeName : "дарган мез", flag : ""},
    day : { englishName : "Dayak", nativeName : "Dayak", flag : ""},
    del : { englishName : "Delaware", nativeName : "Lënape", flag : ""},
    chp : { englishName : "Dene Suline", nativeName : "Dëne Sųłiné", flag : ""},
    din : { englishName : "Dinka", nativeName : "Thuɔŋjäŋ", flag : ""},
    div : { englishName : "Divehi", nativeName : "ދިވެހިބަސ", flag : ""},
    dv : { englishName : "Divehi", nativeName : "ދިވެހިބަސ", flag : ""},
    doi : { englishName : "Dogri", nativeName : "डोगरी", flag : ""},
    dgr : { englishName : "Dogrib", nativeName : "Tłįchǫ", flag : ""},
    dra : { englishName : "Dravidian (Other)", nativeName : "Dravidian (Other)", flag : ""},
    dua : { englishName : "Duala", nativeName : "Duala", flag : ""},
    dut : { englishName : "Dutch", nativeName : "Nederlands", flag : ""},
    nld : { englishName : "Dutch", nativeName : "Nederlands", flag : ""},
    nl : { englishName : "Dutch", nativeName : "Nederlands", flag : ""},
    dum : { englishName : "Dutch, Middle (ca. 1050–1350)", nativeName : "Dutch, Middle (ca. 1050–1350)", flag : ""},
    dyu : { englishName : "Dyula", nativeName : "Julakan", flag : ""},
    dzo : { englishName : "Dzongkha", nativeName : "རྫོང་ཁ", flag : ""},
    dz : { englishName : "Dzongkha", nativeName : "རྫོང་ཁ", flag : ""},
    efi : { englishName : "Efik", nativeName : "Efik", flag : ""},
    egy : { englishName : "Egyptian (Ancient)", nativeName : "Egyptian (Ancient)", flag : ""},
    eka : { englishName : "Ekajuk", nativeName : "Ekajuk", flag : ""},
    elx : { englishName : "Elamite", nativeName : "Elamite", flag : ""},
    enm : { englishName : "English, Middle (ca. 1100–1500)", nativeName : "English, Middle (ca. 1100–1500)", flag : ""},
    ang : { englishName : "English, Old (ca. 450–1100)", nativeName : "Englisc", flag : ""},
    myv : { englishName : "Erzya", nativeName : "эрзянь кель", flag : ""},
    epo : { englishName : "Esperanto", nativeName : "Esperanto", flag : ""},
    eo : { englishName : "Esperanto", nativeName : "Esperanto", flag : ""},
    est : { englishName : "Estonian", nativeName : "eesti keel", flag : ""},
    et : { englishName : "Estonian", nativeName : "eesti keel", flag : ""},
    ewe : { englishName : "Ewe", nativeName : "Ɛʋɛgbɛ", flag : ""},
    ee : { englishName : "Ewe", nativeName : "Ɛʋɛgbɛ", flag : ""},
    ewo : { englishName : "Ewondo", nativeName : "Ewondo", flag : ""},
    fan : { englishName : "Fang", nativeName : "Fang", flag : ""},
    fat : { englishName : "Fanti", nativeName : "Fanti", flag : ""},
    fao : { englishName : "Faroese", nativeName : "føroyskt", flag : ""},
    fo : { englishName : "Faroese", nativeName : "føroyskt", flag : ""},
    fij : { englishName : "Fijian", nativeName : "vosa Vakaviti", flag : ""},
    fj : { englishName : "Fijian", nativeName : "vosa Vakaviti", flag : ""},
    fil : { englishName : "Filipino", nativeName : "Filipino", flag : ""},
    fin : { englishName : "Finnish", nativeName : "suomi", flag : ""},
    fi : { englishName : "Finnish", nativeName : "suomi", flag : ""},
    fiu : { englishName : "Finno-Ugrian (Other)", nativeName : "Finno-Ugrian (Other)", flag : ""},
    fon : { englishName : "Fon", nativeName : "Fɔngbe", flag : ""},
    fre : { englishName : "French", nativeName : "français", flag : ""},
    fra : { englishName : "French", nativeName : "français", flag : ""},
    fr : { englishName : "French", nativeName : "français", flag : ""},
    frm : { englishName : "French, Middle (ca. 1400—1600)", nativeName : "French, Middle (ca. 1400—1600)", flag : ""},
    fro : { englishName : "French, Old (842—ca. 1400)", nativeName : "French, Old (842—ca. 1400)", flag : ""},
    frs : { englishName : "Frisian, Eastern", nativeName : "Seeltersk Fräisk", flag : ""},
    frr : { englishName : "Frisian, Northern", nativeName : "Frisian, Northern", flag : ""},
    fry : { englishName : "Frisian, Western", nativeName : "frysk", flag : ""},
    fy : { englishName : "Frisian, Western", nativeName : "frysk", flag : ""},
    fur : { englishName : "Friulian", nativeName : "furlan", flag : ""},
    ful : { englishName : "Fulah", nativeName : "Fulfulde, Pulaar, Pular", flag : ""},
    ff : { englishName : "Fulah", nativeName : "Fulfulde, Pulaar, Pular", flag : ""},
    gaa : { englishName : "Ga", nativeName : "Gã", flag : ""},
    glg : { englishName : "Galician", nativeName : "Galego", flag : ""},
    gl : { englishName : "Galician", nativeName : "Galego", flag : ""},
    lug : { englishName : "Ganda", nativeName : "Luganda", flag : ""},
    lg : { englishName : "Ganda", nativeName : "Luganda", flag : ""},
    gay : { englishName : "Gayo", nativeName : "Gayo", flag : ""},
    gba : { englishName : "Gbaya", nativeName : "Gbaya", flag : ""},
    gez : { englishName : "Ge'ez", nativeName : "ግዕዝ", flag : ""},
    geo : { englishName : "Georgian", nativeName : "ქართული ენა (kartuli ena)", flag : ""},
    kat : { englishName : "Georgian", nativeName : "ქართული ენა (kartuli ena)", flag : ""},
    ka : { englishName : "Georgian", nativeName : "ქართული ენა (kartuli ena)", flag : ""},
    ger : { englishName : "German", nativeName : "Deutsch", flag : ""},
    deu : { englishName : "German", nativeName : "Deutsch", flag : ""},
    de : { englishName : "German", nativeName : "Deutsch", flag : ""},
    gsw : { englishName : "German, Alemannic", nativeName : "Alemannisch, Schwyzerdütsch", flag : ""},
    nds : { englishName : "German, Low", nativeName : "Nederdüütsch, Plattdüütsch", flag : ""},
    gmh : { englishName : "German, Middle High (ca. 1050–1500)", nativeName : "diutisk", flag : ""},
    goh : { englishName : "German, Old High (ca. 750–1050)", nativeName : "diutisc", flag : ""},
    gem : { englishName : "Germanic (Other)", nativeName : "Germanic (Other)", flag : ""},
    gil : { englishName : "Gilbertese", nativeName : "taetae ni Kiribati", flag : ""},
    gon : { englishName : "Gondi", nativeName : "Gōndi", flag : ""},
    gor : { englishName : "Gorontalo", nativeName : "Gorontalo", flag : ""},
    got : { englishName : "Gothic", nativeName : "𐌲𐌿𐍄𐌹𐍃𐌺", flag : ""},
    grb : { englishName : "Grebo", nativeName : "Grebo", flag : ""},
    grc : { englishName : "Greek, Ancient (to 1453)", nativeName : "Ἑλληνικά", flag : ""},
    gre : { englishName : "Greek, Modern (1453–)", nativeName : "Ελληνικά", flag : ""},
    ell : { englishName : "Greek, Modern (1453–)", nativeName : "Ελληνικά", flag : ""},
    el : { englishName : "Greek, Modern (1453–)", nativeName : "Ελληνικά", flag : ""},
    kal : { englishName : "Greenlandic", nativeName : "kalaallisut", flag : ""},
    kl : { englishName : "Greenlandic", nativeName : "kalaallisut", flag : ""},
    grn : { englishName : "Guarani", nativeName : "Avañe'ẽ", flag : ""},
    gn : { englishName : "Guarani", nativeName : "Avañe'ẽ", flag : ""},
    guj : { englishName : "Gujarati", nativeName : "ગુજરાતી", flag : ""},
    gu : { englishName : "Gujarati", nativeName : "ગુજરાતી", flag : ""},
    gwi : { englishName : "Gwichʼin", nativeName : "Gwichʼin", flag : ""},
    hai : { englishName : "Haida", nativeName : "X̲aat Kíl", flag : ""},
    hat : { englishName : "Haitian Creole", nativeName : "Kreyòl ayisyen", flag : ""},
    ht : { englishName : "Haitian Creole", nativeName : "Kreyòl ayisyen", flag : ""},
    hau : { englishName : "Hausa", nativeName : "Hausancī", flag : ""},
    ha : { englishName : "Hausa", nativeName : "Hausancī", flag : ""},
    haw : { englishName : "Hawaiian", nativeName : "‘Ōlelo Hawai‘i", flag : ""},
    heb : { englishName : "Hebrew", nativeName : "עִבְרִית", flag : ""},
    he : { englishName : "Hebrew", nativeName : "עִבְרִית", flag : ""},
    her : { englishName : "Herero", nativeName : "Otjiherero", flag : ""},
    hz : { englishName : "Herero", nativeName : "Otjiherero", flag : ""},
    hil : { englishName : "Hiligaynon", nativeName : "Ilonggo", flag : ""},
    him : { englishName : "Himachali", nativeName : "Himachali", flag : ""},
    hin : { englishName : "Hindi", nativeName : "हिन्दी", flag : ""},
    hi : { englishName : "Hindi", nativeName : "हिन्दी", flag : ""},
    hmo : { englishName : "Hiri Motu", nativeName : "Hiri Motu", flag : ""},
    ho : { englishName : "Hiri Motu", nativeName : "Hiri Motu", flag : ""},
    hit : { englishName : "Hittite", nativeName : "Hittite", flag : ""},
    hmn : { englishName : "Hmong", nativeName : "Hmoob", flag : ""},
    hun : { englishName : "Hungarian", nativeName : "Magyar", flag : ""},
    hu : { englishName : "Hungarian", nativeName : "Magyar", flag : ""},
    hup : { englishName : "Hupa", nativeName : "Na:tinixwe Mixine:whe", flag : ""},
    iba : { englishName : "Iban", nativeName : "Iban", flag : ""},
    ice : { englishName : "Icelandic", nativeName : "íslenska", flag : ""},
    isl : { englishName : "Icelandic", nativeName : "íslenska", flag : ""},
    is : { englishName : "Icelandic", nativeName : "íslenska", flag : ""},
    ido : { englishName : "Ido", nativeName : "Ido", flag : ""},
    io : { englishName : "Ido", nativeName : "Ido", flag : ""},
    ibo : { englishName : "Igbo", nativeName : "Igbo", flag : ""},
    ig : { englishName : "Igbo", nativeName : "Igbo", flag : ""},
    ijo : { englishName : "Ijo", nativeName : "Ijo", flag : ""},
    ilo : { englishName : "Iloko", nativeName : "Iloko", flag : ""},
    smn : { englishName : "Inari Sami", nativeName : "säämegiella", flag : ""},
    inc : { englishName : "Indic (Other)", nativeName : "Indic (Other)", flag : ""},
    ine : { englishName : "Indo-European (Other)", nativeName : "Indo-European (Other)", flag : ""},
    ind : { englishName : "Indonesian", nativeName : "Bahasa Indonesia", flag : ""},
    id : { englishName : "Indonesian", nativeName : "Bahasa Indonesia", flag : ""},
    inh : { englishName : "Ingush", nativeName : "гӀалгӀай мотт", flag : ""},
    ina : { englishName : "Interlingua (International Auxiliary Language Association)", nativeName : "interlingua", flag : ""},
    ia : { englishName : "Interlingua (International Auxiliary Language Association)", nativeName : "interlingua", flag : ""},
    ile : { englishName : "Interlingue", nativeName : "Interlingue", flag : ""},
    ie : { englishName : "Interlingue", nativeName : "Interlingue", flag : ""},
    iku : { englishName : "Inuktitut", nativeName : "ᐃᓄᒃᑎᑐᑦ", flag : ""},
    iu : { englishName : "Inuktitut", nativeName : "ᐃᓄᒃᑎᑐᑦ", flag : ""},
    ipk : { englishName : "Inupiaq", nativeName : "Iñupiaq", flag : ""},
    ik : { englishName : "Inupiaq", nativeName : "Iñupiaq", flag : ""},
    ira : { englishName : "Iranian (Other)", nativeName : "Iranian (Other)", flag : ""},
    gle : { englishName : "Irish", nativeName : "Gaeilge", flag : ""},
    ga : { englishName : "Irish", nativeName : "Gaeilge", flag : ""},
    mga : { englishName : "Irish, Middle (900–1200)", nativeName : "Gaoidhealg", flag : ""},
    sga : { englishName : "Irish, Old (to 900)", nativeName : "Goídelc", flag : ""},
    iro : { englishName : "Iroquoian languages", nativeName : "Iroquoian languages", flag : ""},
    ita : { englishName : "Italian", nativeName : "italiano", flag : ""},
    it : { englishName : "Italian", nativeName : "italiano", flag : ""},
    jpn : { englishName : "Japanese", nativeName : "日本語", flag : ""},
    ja : { englishName : "Japanese", nativeName : "日本語", flag : ""},
    jav : { englishName : "Javanese", nativeName : "basa Jawa", flag : ""},
    jv : { englishName : "Javanese", nativeName : "basa Jawa", flag : ""},
    jrb : { englishName : "Judeo-Arabic", nativeName : "Judeo-Arabic", flag : ""},
    jpr : { englishName : "Judeo-Persian", nativeName : "Judeo-Persian", flag : ""},
    kbd : { englishName : "Kabardian", nativeName : "къэбэрдеибзэ", flag : ""},
    kab : { englishName : "Kabyle", nativeName : "Taqbaylit", flag : ""},
    kac : { englishName : "Kachin", nativeName : "Jingpho", flag : ""},
    xal : { englishName : "Kalmyk", nativeName : "хальмг келн", flag : ""},
    kam : { englishName : "Kamba", nativeName : "Kamba", flag : ""},
    kan : { englishName : "Kannada", nativeName : "ಕನ್ನಡ", flag : ""},
    kn : { englishName : "Kannada", nativeName : "ಕನ್ನಡ", flag : ""},
    kau : { englishName : "Kanuri", nativeName : "Kanuri", flag : ""},
    kr : { englishName : "Kanuri", nativeName : "Kanuri", flag : ""},
    krc : { englishName : "Karachay-Balkar", nativeName : "къарачай-малкъар тил", flag : ""},
    kaa : { englishName : "Kara-Kalpak", nativeName : "қарақалпақ тили", flag : ""},
    krl : { englishName : "Karelian", nativeName : "karjalan kieli", flag : ""},
    kar : { englishName : "Karen", nativeName : "Karen", flag : ""},
    kas : { englishName : "Kashmiri", nativeName : "कॉशुर", flag : ""},
    ks : { englishName : "Kashmiri", nativeName : "कॉशुर", flag : ""},
    csb : { englishName : "Kashubian", nativeName : "kaszëbsczi jãzëk", flag : ""},
    kaw : { englishName : "Kawi", nativeName : "Bhāṣa Kawi", flag : ""},
    kaz : { englishName : "Kazakh", nativeName : "Қазақ тілі", flag : ""},
    kk : { englishName : "Kazakh", nativeName : "Қазақ тілі", flag : ""},
    kha : { englishName : "Khasi", nativeName : "Khasi", flag : ""},
    khm : { englishName : "Khmer", nativeName : "Khmer", flag : ""},
    km : { englishName : "Khmer", nativeName : "Khmer", flag : ""},
    khi : { englishName : "Khoisan (Other)", nativeName : "Khoisan (Other)", flag : ""},
    kho : { englishName : "Khotanese", nativeName : "Khotanese", flag : ""},
    kik : { englishName : "Kikuyu", nativeName : "Gĩkũyũ", flag : ""},
    ki : { englishName : "Kikuyu", nativeName : "Gĩkũyũ", flag : ""},
    kmb : { englishName : "Kimbundu", nativeName : "Kimbundu", flag : ""},
    kin : { englishName : "Kinyarwanda", nativeName : "kinyaRwanda", flag : ""},
    rw : { englishName : "Kinyarwanda", nativeName : "kinyaRwanda", flag : ""},
    kir : { englishName : "Kirghiz", nativeName : "кыргыз тили", flag : ""},
    ky : { englishName : "Kirghiz", nativeName : "кыргыз тили", flag : ""},
    tlh : { englishName : "Klingon", nativeName : "tlhIngan Hol", flag : ""},
    kom : { englishName : "Komi", nativeName : "коми кыв", flag : ""},
    kv : { englishName : "Komi", nativeName : "коми кыв", flag : ""},
    kon : { englishName : "Kongo", nativeName : "Kikongo", flag : ""},
    kg : { englishName : "Kongo", nativeName : "Kikongo", flag : ""},
    kok : { englishName : "Konkani", nativeName : "कोंकणी", flag : ""},
    kor : { englishName : "Korean", nativeName : "한국어", flag : ""},
    ko : { englishName : "Korean", nativeName : "한국어", flag : ""},
    kos : { englishName : "Kosraean", nativeName : "Kosrae", flag : ""},
    kpe : { englishName : "Kpelle", nativeName : "kpele", flag : ""},
    kro : { englishName : "Kru", nativeName : "Kru", flag : ""},
    kua : { englishName : "Kuanyama", nativeName : "Kuanyama", flag : ""},
    kj : { englishName : "Kuanyama", nativeName : "Kuanyama", flag : ""},
    kum : { englishName : "Kumyk", nativeName : "Кумык", flag : ""},
    kur : { englishName : "Kurdish", nativeName : "Kurdî", flag : ""},
    ku : { englishName : "Kurdish", nativeName : "Kurdî", flag : ""},
    kru : { englishName : "Kurukh", nativeName : "Kurukh", flag : ""},
    kut : { englishName : "Kutenai", nativeName : "Ktunaxa", flag : ""},
    lad : { englishName : "Ladino", nativeName : "ג'ודיאו-איספאנייול", flag : ""},
    lah : { englishName : "Lahnda", nativeName : "ਲਹਿੰਦੀ", flag : ""},
    lam : { englishName : "Lamba", nativeName : "Lamba", flag : ""},
    lao : { englishName : "Lao", nativeName : "ພາສາລາວ", flag : ""},
    lo : { englishName : "Lao", nativeName : "ພາສາລາວ", flag : ""},
    lat : { englishName : "Latin", nativeName : "latine", flag : ""},
    la : { englishName : "Latin", nativeName : "latine", flag : ""},
    lav : { englishName : "Latvian", nativeName : "latviešu valoda", flag : ""},
    lv : { englishName : "Latvian", nativeName : "latviešu valoda", flag : ""},
    lez : { englishName : "Lezghian", nativeName : "лезги чӀал", flag : ""},
    lim : { englishName : "Limburgish", nativeName : "Limburgs", flag : ""},
    li : { englishName : "Limburgish", nativeName : "Limburgs", flag : ""},
    lin : { englishName : "Lingala", nativeName : "lingala", flag : ""},
    ln : { englishName : "Lingala", nativeName : "lingala", flag : ""},
    lit : { englishName : "Lithuanian", nativeName : "lietuvių kalba", flag : ""},
    lt : { englishName : "Lithuanian", nativeName : "lietuvių kalba", flag : ""},
    jbo : { englishName : "Lojban", nativeName : "lojban", flag : ""},
    loz : { englishName : "Lozi", nativeName : "siLozi", flag : ""},
    lub : { englishName : "Luba-Katanga", nativeName : "Luba-Katanga", flag : ""},
    lu : { englishName : "Luba-Katanga", nativeName : "Luba-Katanga", flag : ""},
    lua : { englishName : "Luba-Lulua", nativeName : "lwaà:", flag : ""},
    lui : { englishName : "Luiseño", nativeName : "Luiseño", flag : ""},
    smj : { englishName : "Lule Sami", nativeName : "sámegiella", flag : ""},
    lun : { englishName : "Lunda", nativeName : "chiLunda", flag : ""},
    luo : { englishName : "Luo (Kenya and Tanzania)", nativeName : "Dholuo", flag : ""},
    lus : { englishName : "Lushai", nativeName : "Lushai", flag : ""},
    ltz : { englishName : "Luxembourgish", nativeName : "Lëtzebuergesch", flag : ""},
    lb : { englishName : "Luxembourgish", nativeName : "Lëtzebuergesch", flag : ""},
    mac : { englishName : "Macedonian", nativeName : "македонски јазик", flag : ""},
    mkd : { englishName : "Macedonian", nativeName : "македонски јазик", flag : ""},
    mk : { englishName : "Macedonian", nativeName : "македонски јазик", flag : ""},
    mad : { englishName : "Madurese", nativeName : "Madurese", flag : ""},
    mag : { englishName : "Magahi", nativeName : "Magahi", flag : ""},
    mai : { englishName : "Maithili", nativeName : "मैथिली", flag : ""},
    mak : { englishName : "Makasar", nativeName : "Makasar", flag : ""},
    mlg : { englishName : "Malagasy", nativeName : "Malagasy fiteny", flag : ""},
    mg : { englishName : "Malagasy", nativeName : "Malagasy fiteny", flag : ""},
    may : { englishName : "Malay", nativeName : "bahasa Melayu", flag : ""},
    msa : { englishName : "Malay", nativeName : "bahasa Melayu", flag : ""},
    ms : { englishName : "Malay", nativeName : "bahasa Melayu", flag : ""},
    mal : { englishName : "Malayalam", nativeName : "മലയാളം", flag : ""},
    ml : { englishName : "Malayalam", nativeName : "മലയാളം", flag : ""},
    mlt : { englishName : "Maltese", nativeName : "Malti", flag : ""},
    mt : { englishName : "Maltese", nativeName : "Malti", flag : ""},
    mnc : { englishName : "Manchu", nativeName : "ᠮᠠᠨᠵᡠ ᡤᡳᠰᡠᠨ ᠪᡝ", flag : ""},
    mdr : { englishName : "Mandar", nativeName : "Mandar", flag : ""},
    man : { englishName : "Mandingo", nativeName : "Mandingo", flag : ""},
    mni : { englishName : "Manipuri", nativeName : "মৈইতৈইলোন", flag : ""},
    mno : { englishName : "Manobo languages", nativeName : "Manobo languages", flag : ""},
    glv : { englishName : "Manx", nativeName : "Gaelg", flag : ""},
    gv : { englishName : "Manx", nativeName : "Gaelg", flag : ""},
    mao : { englishName : "Māori", nativeName : "te reo Māori", flag : ""},
    mri : { englishName : "Māori", nativeName : "te reo Māori", flag : ""},
    mi : { englishName : "Māori", nativeName : "te reo Māori", flag : ""},
    mar : { englishName : "Marathi", nativeName : "मराठी", flag : ""},
    mr : { englishName : "Marathi", nativeName : "मराठी", flag : ""},
    chm : { englishName : "Mari", nativeName : "марий йылме", flag : ""},
    mah : { englishName : "Marshallese", nativeName : "Kajin M̧ajeļ", flag : ""},
    mh : { englishName : "Marshallese", nativeName : "Kajin M̧ajeļ", flag : ""},
    mwr : { englishName : "Marwari", nativeName : "मारवाड़ी", flag : ""},
    mas : { englishName : "Masai", nativeName : "ɔl Maa", flag : ""},
    myn : { englishName : "Mayan languages", nativeName : "Mayan languages", flag : ""},
    men : { englishName : "Mende", nativeName : "Mɛnde", flag : ""},
    mic : { englishName : "Mi'kmaq", nativeName : "Míkmaq, Mi'gmaq", flag : ""},
    min : { englishName : "Minangkabau", nativeName : "Baso Minangkabau", flag : ""},
    mwl : { englishName : "Mirandese", nativeName : "Lhéngua Mirandesa", flag : ""},
    moh : { englishName : "Mohawk", nativeName : "Kanien’keha", flag : ""},
    mdf : { englishName : "Moksha", nativeName : "мокшень кяль", flag : ""},
    mol : { englishName : "Moldavian", nativeName : "лимба молдовеняскэ", flag : ""},
    mo : { englishName : "Moldavian", nativeName : "лимба молдовеняскэ", flag : ""},
    mkh : { englishName : "Mon-Khmer (Other)", nativeName : "Mon-Khmer (Other)", flag : ""},
    lol : { englishName : "Mongo", nativeName : "Mongo", flag : ""},
    mon : { englishName : "Mongolian", nativeName : "монгол хэл", flag : ""},
    mn : { englishName : "Mongolian", nativeName : "монгол хэл", flag : ""},
    mos : { englishName : "Mossi", nativeName : "Mòoré", flag : ""},
    mun : { englishName : "Munda languages", nativeName : "Munda languages", flag : ""},
    nah : { englishName : "Nahuatl", nativeName : "nāhuatl", flag : ""},
    nau : { englishName : "Nauruan", nativeName : "Ekakairũ Naoero", flag : ""},
    na : { englishName : "Nauruan", nativeName : "Ekakairũ Naoero", flag : ""},
    nav : { englishName : "Navajo", nativeName : "Diné bizaad", flag : ""},
    nv : { englishName : "Navajo", nativeName : "Diné bizaad", flag : ""},
    nde : { englishName : "Ndebele, North", nativeName : "isiNdebele", flag : ""},
    nd : { englishName : "Ndebele, North", nativeName : "isiNdebele", flag : ""},
    nbl : { englishName : "Ndebele, South", nativeName : "Ndébélé", flag : ""},
    nr : { englishName : "Ndebele, South", nativeName : "Ndébélé", flag : ""},
    ndo : { englishName : "Ndonga", nativeName : "Owambo", flag : ""},
    ng : { englishName : "Ndonga", nativeName : "Owambo", flag : ""},
    nap : { englishName : "Neapolitan", nativeName : "napulitano", flag : ""},
    new : { englishName : "Nepal Bhasa", nativeName : "Nepal Bhasa", flag : ""},
    nep : { englishName : "Nepali", nativeName : "नेपाली", flag : ""},
    ne : { englishName : "Nepali", nativeName : "नेपाली", flag : ""},
    nia : { englishName : "Nias", nativeName : "Nias", flag : ""},
    nic : { englishName : "Niger-Kordofanian (Other)", nativeName : "Niger-Kordofanian (Other)", flag : ""},
    ssa : { englishName : "Nilo-Saharan (Other)", nativeName : "Nilo-Saharan (Other)", flag : ""},
    niu : { englishName : "Niuean", nativeName : "ko e vagahau Niuē", flag : ""},
    nqo : { englishName : "N'Ko", nativeName : "N'Ko", flag : ""},
    nog : { englishName : "Nogai", nativeName : "ногай тили", flag : ""},
    non : { englishName : "Norse, Old", nativeName : "norskr", flag : ""},
    nai : { englishName : "North American Indian (Other)", nativeName : "North American Indian (Other)", flag : ""},
    sme : { englishName : "Northern Sami", nativeName : "sámi", flag : ""},
    se : { englishName : "Northern Sami", nativeName : "sámi", flag : ""},
    nor : { englishName : "Norwegian", nativeName : "Norsk", flag : ""},
    no : { englishName : "Norwegian", nativeName : "Norsk", flag : ""},
    nob : { englishName : "Norwegian Bokmål", nativeName : "Norsk bokmål", flag : ""},
    nb : { englishName : "Norwegian Bokmål", nativeName : "Norsk bokmål", flag : ""},
    nno : { englishName : "Norwegian Nynorsk", nativeName : "Norsk nynorsk", flag : ""},
    nn : { englishName : "Norwegian Nynorsk", nativeName : "Norsk nynorsk", flag : ""},
    nub : { englishName : "Nubian languages", nativeName : "Nubian languages", flag : ""},
    nym : { englishName : "Nyamwezi", nativeName : "Kinyamwezi", flag : ""},
    nyn : { englishName : "Nyankole", nativeName : "Nyankole", flag : ""},
    nyo : { englishName : "Nyoro", nativeName : "Runyoro", flag : ""},
    nzi : { englishName : "Nzima", nativeName : "Nzima", flag : ""},
    oci : { englishName : "Occitan (post 1500)", nativeName : "Occitan", flag : ""},
    oc : { englishName : "Occitan (post 1500)", nativeName : "Occitan", flag : ""},
    oji : { englishName : "Ojibwa, Anishinaabe languages", nativeName : "ᐊᓂᔑᓇᐯᒧᐏᐣ (Anishinaabemowin)", flag : ""},
    oj : { englishName : "Ojibwa, Anishinaabe languages", nativeName : "ᐊᓂᔑᓇᐯᒧᐏᐣ (Anishinaabemowin)", flag : ""},
    ori : { englishName : "Oriya", nativeName : "ଓଡ଼ିଆ", flag : ""},
    or : { englishName : "Oriya", nativeName : "ଓଡ଼ିଆ", flag : ""},
    orm : { englishName : "Oromo", nativeName : "Afaan Oromoo", flag : ""},
    om : { englishName : "Oromo", nativeName : "Afaan Oromoo", flag : ""},
    osa : { englishName : "Osage", nativeName : "Osage", flag : ""},
    oss : { englishName : "Ossetian", nativeName : "ирон ӕвзаг", flag : ""},
    os : { englishName : "Ossetian", nativeName : "ирон ӕвзаг", flag : ""},
    oto : { englishName : "Otomian languages", nativeName : "Otomian languages", flag : ""},
    pal : { englishName : "Pahlavi (Middle Persian)", nativeName : "Pahlavi (Middle Persian)", flag : ""},
    pau : { englishName : "Palauan", nativeName : "tekoi ra Belau", flag : ""},
    pli : { englishName : "Pali", nativeName : "पालि", flag : ""},
    pi : { englishName : "Pali", nativeName : "पालि", flag : ""},
    pam : { englishName : "Pampanga", nativeName : "Kapampangan", flag : ""},
    pag : { englishName : "Pangasinan", nativeName : "Pangasinan", flag : ""},
    pap : { englishName : "Papiamento", nativeName : "Papiamentu", flag : ""},
    paa : { englishName : "Papuan (Other)", nativeName : "Papuan (Other)", flag : ""},
    per : { englishName : "Persian", nativeName : "فارسی", flag : ""},
    fas : { englishName : "Persian", nativeName : "فارسی", flag : ""},
    fa : { englishName : "Persian", nativeName : "فارسی", flag : ""},
    peo : { englishName : "Persian, Old (ca. 600–400 BC)", nativeName : "Persian, Old (ca. 600–400 BC)", flag : ""},
    phi : { englishName : "Philippine (Other)", nativeName : "Philippine (Other)", flag : ""},
    phn : { englishName : "Phoenician", nativeName : "Phoenician", flag : ""},
    pon : { englishName : "Pohnpeian", nativeName : "Pohnpeian", flag : ""},
    pol : { englishName : "Polish", nativeName : "polski", flag : ""},
    pl : { englishName : "Polish", nativeName : "polski", flag : ""},
    por : { englishName : "Portuguese", nativeName : "português", flag : ""},
    pt : { englishName : "Portuguese", nativeName : "português", flag : ""},
    pra : { englishName : "Prakrit languages", nativeName : "Prakrit languages", flag : ""},
    pro : { englishName : "Provençal, Old (to 1500)", nativeName : "Provençal, Old (to 1500)", flag : ""},
    pan : { englishName : "Punjabi", nativeName : "ਪੰਜਾਬੀ", flag : ""},
    pa : { englishName : "Punjabi", nativeName : "ਪੰਜਾਬੀ", flag : ""},
    pus : { englishName : "Pushto", nativeName : "پښت", flag : ""},
    ps : { englishName : "Pushto", nativeName : "پښت", flag : ""},
    que : { englishName : "Quechuan languages", nativeName : "Runa Simi", flag : ""},
    qu : { englishName : "Quechuan languages", nativeName : "Runa Simi", flag : ""},
    roh : { englishName : "Raeto-Romance", nativeName : "rumantsch grischun", flag : ""},
    rm : { englishName : "Raeto-Romance", nativeName : "rumantsch grischun", flag : ""},
    raj : { englishName : "Rajasthani", nativeName : "राजस्थानी", flag : ""},
    rap : { englishName : "Rapanui", nativeName : "rapanui", flag : ""},
    rar : { englishName : "Rarotongan", nativeName : "Rarotongan", flag : ""},
    roa : { englishName : "Romance (Other)", nativeName : "Romance (Other)", flag : ""},
    rum : { englishName : "Romanian", nativeName : "română", flag : ""},
    ron : { englishName : "Romanian", nativeName : "română", flag : ""},
    ro : { englishName : "Romanian", nativeName : "română", flag : ""},
    rom : { englishName : "Romany", nativeName : "rromani ćhib", flag : ""},
    run : { englishName : "Rundi", nativeName : "kiRundi", flag : ""},
    rn : { englishName : "Rundi", nativeName : "kiRundi", flag : ""},
    rus : { englishName : "Russian", nativeName : "русский язык", flag : ""},
    ru : { englishName : "Russian", nativeName : "русский язык", flag : ""},
    sal : { englishName : "Salishan languages", nativeName : "Salishan languages", flag : ""},
    sam : { englishName : "Samaritan Aramaic", nativeName : "ארמית, ܐܪܡܝܐ", flag : ""},
    smi : { englishName : "Sami languages (Other)", nativeName : "Sami languages (Other)", flag : ""},
    smo : { englishName : "Samoan", nativeName : "gagana fa'a Samoa", flag : ""},
    sm : { englishName : "Samoan", nativeName : "gagana fa'a Samoa", flag : ""},
    sad : { englishName : "Sandawe", nativeName : "Sandawe", flag : ""},
    sag : { englishName : "Sango", nativeName : "yângâ tî sängö", flag : ""},
    sg : { englishName : "Sango", nativeName : "yângâ tî sängö", flag : ""},
    san : { englishName : "Sanskrit", nativeName : "संस्कृतम्", flag : ""},
    sa : { englishName : "Sanskrit", nativeName : "संस्कृतम्", flag : ""},
    sat : { englishName : "Santali", nativeName : "संथाली", flag : ""},
    srd : { englishName : "Sardinian", nativeName : "sardu", flag : ""},
    sc : { englishName : "Sardinian", nativeName : "sardu", flag : ""},
    sas : { englishName : "Sasak", nativeName : "Sasak", flag : ""},
    sco : { englishName : "Scots", nativeName : "Scoats leid", flag : ""},
    gla : { englishName : "Scottish Gaelic", nativeName : "Gàidhlig", flag : ""},
    gd : { englishName : "Scottish Gaelic", nativeName : "Gàidhlig", flag : ""},
    sel : { englishName : "Selkup", nativeName : "шӧльӄумыт әты", flag : ""},
    sem : { englishName : "Semitic (Other)", nativeName : "Semitic (Other)", flag : ""},
    scc : { englishName : "Serbian", nativeName : "Српски / Srpski", flag : ""},
    srp : { englishName : "Serbian", nativeName : "Српски / Srpski", flag : ""},
    sr : { englishName : "Serbian", nativeName : "Српски / Srpski", flag : ""},
    srr : { englishName : "Serer", nativeName : "Serer", flag : ""},
    shn : { englishName : "Shan", nativeName : "Shan", flag : ""},
    sna : { englishName : "Shona", nativeName : "chiShona", flag : ""},
    sn : { englishName : "Shona", nativeName : "chiShona", flag : ""},
    iii : { englishName : "Sichuan Yi", nativeName : "ꆇꉙ", flag : ""},
    ii : { englishName : "Sichuan Yi", nativeName : "ꆇꉙ", flag : ""},
    scn : { englishName : "Sicilian", nativeName : "Sicilianu", flag : ""},
    sid : { englishName : "Sidamo", nativeName : "Sidámo 'Afó", flag : ""},
    sgn : { englishName : "Sign languages", nativeName : "Sign languages", flag : ""},
    bla : { englishName : "Siksika", nativeName : "siksiká", flag : ""},
    snd : { englishName : "Sindhi", nativeName : "سنڌي، سندھی, सिन्धी", flag : ""},
    sd : { englishName : "Sindhi", nativeName : "سنڌي، سندھی, सिन्धी", flag : ""},
    sin : { englishName : "Sinhalese", nativeName : "සිංහල", flag : ""},
    si : { englishName : "Sinhalese", nativeName : "සිංහල", flag : ""},
    sit : { englishName : "Sino-Tibetan (Other)", nativeName : "Sino-Tibetan (Other)", flag : ""},
    sio : { englishName : "Siouan languages", nativeName : "Siouan languages", flag : ""},
    sms : { englishName : "Skolt Sami", nativeName : "sääʼmǩiõll", flag : ""},
    den : { englishName : "Slave (Athapascan)", nativeName : "Slave (Athapascan)", flag : ""},
    sla : { englishName : "Slavic (Other)", nativeName : "Slavic (Other)", flag : ""},
    slo : { englishName : "Slovak", nativeName : "slovenčina", flag : ""},
    slk : { englishName : "Slovak", nativeName : "slovenčina", flag : ""},
    sk : { englishName : "Slovak", nativeName : "slovenčina", flag : ""},
    slv : { englishName : "Slovenian", nativeName : "slovenščina", flag : ""},
    sl : { englishName : "Slovenian", nativeName : "slovenščina", flag : ""},
    sog : { englishName : "Sogdian", nativeName : "Sogdian", flag : ""},
    som : { englishName : "Somali", nativeName : "Soomaaliga", flag : ""},
    so : { englishName : "Somali", nativeName : "Soomaaliga", flag : ""},
    son : { englishName : "Songhai", nativeName : "Songhai", flag : ""},
    snk : { englishName : "Soninke", nativeName : "Soninkanxaane", flag : ""},
    wen : { englishName : "Sorbian languages", nativeName : "Sorbian languages", flag : ""},
    dsb : { englishName : "Sorbian, Lower", nativeName : "dolnoserbski", flag : ""},
    hsb : { englishName : "Sorbian, Upper", nativeName : "hornjoserbsce", flag : ""},
    nso : { englishName : "Sotho, Northern", nativeName : "sePêdi", flag : ""},
    sot : { englishName : "Sotho, Southern", nativeName : "seSotho", flag : ""},
    st : { englishName : "Sotho, Southern", nativeName : "seSotho", flag : ""},
    sai : { englishName : "South American Indian (Other)", nativeName : "South American Indian (Other)", flag : ""},
    alt : { englishName : "Southern Altai", nativeName : "алтай тили", flag : ""},
    sma : { englishName : "Southern Sami", nativeName : "saemien giele", flag : ""},
    spa : { englishName : "Spanish", nativeName : "español, castellano", flag : ""},
    es : { englishName : "Spanish", nativeName : "español, castellano", flag : ""},
    srn : { englishName : "Sranan Tongo", nativeName : "Sranan Tongo", flag : ""},
    suk : { englishName : "Sukuma", nativeName : "Sukuma", flag : ""},
    sux : { englishName : "Sumerian", nativeName : "eme-ĝir", flag : ""},
    sun : { englishName : "Sundanese", nativeName : "basa Sunda", flag : ""},
    su : { englishName : "Sundanese", nativeName : "basa Sunda", flag : ""},
    sus : { englishName : "Susu", nativeName : "Susu", flag : ""},
    swa : { englishName : "Swahili", nativeName : "kiswahili", flag : ""},
    sw : { englishName : "Swahili", nativeName : "kiswahili", flag : ""},
    ssw : { englishName : "Swati", nativeName : "siSwati", flag : ""},
    ss : { englishName : "Swati", nativeName : "siSwati", flag : ""},
    swe : { englishName : "Swedish", nativeName : "Svenska", flag : ""},
    sv : { englishName : "Swedish", nativeName : "Svenska", flag : ""},
    syr : { englishName : "Syriac", nativeName : "ܣܘܪܝܝܐ", flag : ""},
    tgl : { englishName : "Tagalog", nativeName : "Tagalog", flag : ""},
    tl : { englishName : "Tagalog", nativeName : "Tagalog", flag : ""},
    tah : { englishName : "Tahitian", nativeName : "te reo Tahiti", flag : ""},
    ty : { englishName : "Tahitian", nativeName : "te reo Tahiti", flag : ""},
    tai : { englishName : "Tai (Other)", nativeName : "Tai (Other)", flag : ""},
    tgk : { englishName : "Tajik", nativeName : "тоҷикӣ", flag : ""},
    tg : { englishName : "Tajik", nativeName : "тоҷикӣ", flag : ""},
    tmh : { englishName : "Tamashek", nativeName : "Tamajeq, ɫ[ǂ…", flag : ""},
    tam : { englishName : "Tamil", nativeName : "தமிழ்", flag : ""},
    ta : { englishName : "Tamil", nativeName : "தமிழ்", flag : ""},
    tat : { englishName : "Tatar", nativeName : "татарча", flag : ""},
    tt : { englishName : "Tatar", nativeName : "татарча", flag : ""},
    tel : { englishName : "Telugu", nativeName : "తెలుగు", flag : ""},
    te : { englishName : "Telugu", nativeName : "తెలుగు", flag : ""},
    ter : { englishName : "Tereno", nativeName : "Tereno", flag : ""},
    tet : { englishName : "Tetum, Lia-Tetun", nativeName : "Tetun", flag : ""},
    tha : { englishName : "Thai", nativeName : "ภาษาไทย", flag : ""},
    th : { englishName : "Thai", nativeName : "ภาษาไทย", flag : ""},
    tib : { englishName : "Tibetan", nativeName : "བོད་ཡིག", flag : ""},
    bod : { englishName : "Tibetan", nativeName : "བོད་ཡིག", flag : ""},
    bo : { englishName : "Tibetan", nativeName : "བོད་ཡིག", flag : ""},
    tig : { englishName : "Tigre", nativeName : "Tigré", flag : ""},
    tir : { englishName : "Tigrinya", nativeName : "ትግርኛ", flag : ""},
    ti : { englishName : "Tigrinya", nativeName : "ትግርኛ", flag : ""},
    tem : { englishName : "Timne", nativeName : "Timne", flag : ""},
    tiv : { englishName : "Tiv", nativeName : "Tiv", flag : ""},
    tli : { englishName : "Tlingit", nativeName : "Lingít", flag : ""},
    tpi : { englishName : "Tok Pisin", nativeName : "Tok Pisin", flag : ""},
    tkl : { englishName : "Tokelau", nativeName : "Tokelau", flag : ""},
    tog : { englishName : "Tonga (Malawi)", nativeName : "chiTonga", flag : ""},
    ton : { englishName : "Tongan", nativeName : "faka-Tonga", flag : ""},
    to : { englishName : "Tongan", nativeName : "faka-Tonga", flag : ""},
    tsi : { englishName : "Tsimshian", nativeName : "Tsimshian", flag : ""},
    tso : { englishName : "Tsonga", nativeName : "xiTsonga", flag : ""},
    ts : { englishName : "Tsonga", nativeName : "xiTsonga", flag : ""},
    tsn : { englishName : "Tswana", nativeName : "seTswana", flag : ""},
    tn : { englishName : "Tswana", nativeName : "seTswana", flag : ""},
    tum : { englishName : "Tumbuka", nativeName : "chiTumbuka", flag : ""},
    tup : { englishName : "Tupi languages", nativeName : "Nheengatu", flag : ""},
    tur : { englishName : "Turkish", nativeName : "Türkçe", flag : ""},
    tr : { englishName : "Turkish", nativeName : "Türkçe", flag : ""},
    ota : { englishName : "Turkish, Ottoman (1500–1928)", nativeName : "Turkish, Ottoman (1500–1928)", flag : ""},
    tuk : { englishName : "Turkmen", nativeName : "Түркмен", flag : ""},
    tk : { englishName : "Turkmen", nativeName : "Түркмен", flag : ""},
    tvl : { englishName : "Tuvalu", nativeName : "'gana Tuvalu", flag : ""},
    tyv : { englishName : "Tuvinian", nativeName : "тыва дыл", flag : ""},
    twi : { englishName : "Twi", nativeName : "Twi", flag : ""},
    tw : { englishName : "Twi", nativeName : "Twi", flag : ""},
    udm : { englishName : "Udmurt", nativeName : "удмурт кыл", flag : ""},
    uga : { englishName : "Ugaritic", nativeName : "Ugaritic", flag : ""},
    uig : { englishName : "Uighur", nativeName : "Uyƣurqə", flag : ""},
    ug : { englishName : "Uighur", nativeName : "Uyƣurqə", flag : ""},
    ukr : { englishName : "Ukrainian", nativeName : "українська мова", flag : ""},
    uk : { englishName : "Ukrainian", nativeName : "українська мова", flag : ""},
    umb : { englishName : "Umbundu", nativeName : "úmbúndú", flag : ""},
    urd : { englishName : "Urdu", nativeName : "اردو", flag : ""},
    ur : { englishName : "Urdu", nativeName : "اردو", flag : ""},
    uzb : { englishName : "Uzbek", nativeName : "O'zbek", flag : ""},
    uz : { englishName : "Uzbek", nativeName : "O'zbek", flag : ""},
    vai : { englishName : "Vai", nativeName : "Vai", flag : ""},
    ven : { englishName : "Venda", nativeName : "tshiVenḓa", flag : ""},
    ve : { englishName : "Venda", nativeName : "tshiVenḓa", flag : ""},
    vie : { englishName : "Vietnamese", nativeName : "Tiếng Việt", flag : ""},
    vi : { englishName : "Vietnamese", nativeName : "Tiếng Việt", flag : ""},
    vol : { englishName : "Volapük", nativeName : "Volapük", flag : ""},
    vo : { englishName : "Volapük", nativeName : "Volapük", flag : ""},
    vot : { englishName : "Votic", nativeName : "vaďďa tšeeli", flag : ""},
    wak : { englishName : "Wakashan languages", nativeName : "Wakashan languages", flag : ""},
    wal : { englishName : "Walamo", nativeName : "Walamo", flag : ""},
    wln : { englishName : "Walloon", nativeName : "walon", flag : ""},
    wa : { englishName : "Walloon", nativeName : "walon", flag : ""},
    war : { englishName : "Waray", nativeName : "Winaray", flag : ""},
    was : { englishName : "Washo", nativeName : "Washo", flag : ""},
    wel : { englishName : "Welsh", nativeName : "Cymraeg", flag : ""},
    cym : { englishName : "Welsh", nativeName : "Cymraeg", flag : ""},
    cy : { englishName : "Welsh", nativeName : "Cymraeg", flag : ""},
    wol : { englishName : "Wolof", nativeName : "Wolof", flag : ""},
    wo : { englishName : "Wolof", nativeName : "Wolof", flag : ""},
    xho : { englishName : "Xhosa", nativeName : "isiXhosa", flag : ""},
    xh : { englishName : "Xhosa", nativeName : "isiXhosa", flag : ""},
    sah : { englishName : "Yakut", nativeName : "Саха тыла", flag : ""},
    yao : { englishName : "Yao", nativeName : "Chiyao", flag : ""},
    yap : { englishName : "Yapese", nativeName : "Yapese", flag : ""},
    yid : { englishName : "Yiddish", nativeName : "ייִדיש", flag : ""},
    yi : { englishName : "Yiddish", nativeName : "ייִדיש", flag : ""},
    yor : { englishName : "Yoruba", nativeName : "Yorùbá", flag : ""},
    yo : { englishName : "Yoruba", nativeName : "Yorùbá", flag : ""},
    ypk : { englishName : "Yupik languages", nativeName : "Yupik languages", flag : ""},
    znd : { englishName : "Zande", nativeName : "Pazande", flag : ""},
    zza : { englishName : "Zaza", nativeName : "Zaza", flag : ""},
    zap : { englishName : "Zapotec", nativeName : "Zapotec", flag : ""},
    zen : { englishName : "Zenaga", nativeName : "Tuḍḍungiyya", flag : ""},
    zha : { englishName : "Zhuang", nativeName : "Saɯ cueŋƅ", flag : ""},
    za : { englishName : "Zhuang", nativeName : "Saɯ cueŋƅ", flag : ""},
    zul : { englishName : "Zulu", nativeName : "isiZulu", flag : ""},
    zu : { englishName : "Zulu", nativeName : "isiZulu", flag : ""},
    zun : { englishName : "Zuni", nativeName : "Shiwi", flag : ""}
},


sslib.Map.prototype = {
    clear : function() {
        this.aBuckets = [];
    },

    bucket : function(obj) {
        var i = this.fHash(obj) % this.size;
        if (!this.aBuckets[i])
            this.aBuckets[i] = [];
        return this.aBuckets[i];
    },

    foreach : function(fAction) {
        for each (var bucket in this.aBuckets) {
            for each (var entry in bucket) {
                fAction(entry.key, entry.value);
            }
        }
    },

    contains : function(key) {
        return this.findInBucket(this.bucket(key), key);
    },

    findInBucket : function(bucket, key) {
        for each (var entry in bucket) {
            if (this.fEquals(key, entry.key))
                return entry;
        }
        return null;
    },

    findPositionInBucket : function(bucket, key) {
        for (var i in bucket) {
            var entry = bucket[i];
            if (this.fEquals(key, entry.key))
                return i;
        }
        return -1;
    },

    get : function(key) {
        var bucket = this.bucket(key);
        var entry = this.findInBucket(bucket, key);
        if (entry) return entry.value;
    },

    // Encapsulate the mapped value in an object with one field
    // 'value', or return null.  This allows the caller to distinguish
    // between the key being missing and the key mapping explicitly to
    // null.
    getNull : function(key) {
        var bucket = this.bucket(key);
        var entry = this.findInBucket(bucket, key);
        if (entry)
            return { value : entry.value };
    },

    put : function(key, value) {
        if (typeof value == "undefined")
            return this.remove(key);
        var bucket = this.bucket(key);
        var entry = this.findInBucket(bucket, key);
        if (entry) {
            var old = entry.value;
            entry.value = value;
            return old;
        }

        entry = { key : key, value : value };
        bucket.push(entry);
    },

    remove : function(key) {
        var bucket = this.bucket(key);
        var i = this.findPositionInBucket(bucket, key);
        if (i < 0)
            return;
        var entry = bucket[i];
        bucket.splice(i, 1);
        return entry.value;
    }
};

sslib.initialize();
sslib.debug('Library installed');

var Application = Components.classes["@mozilla.org/fuel/application;1"]
    .getService(Components.interfaces.fuelIApplication);

Application.storage.set("org.standard-sitemap.library", sslib);
